#!/usr/bin/perl -w
#
#  cofolga2mo.pl: A perl script for performing pairwise RNA sequence
#                 alignment by Cofolga2mo.
#
#  Usage:
#  cofolga2mo.pl [fasta file] [--I/--P/--s] [--n [max. # output alignments]]
#
#  '--n 0' indicates output all alignments.
#
# version 0.2.0: linear weight index is available.:

$os = "linux" ;
#$os = "win" ;

if ($os eq "linux") {
   # linux
   $exe = 'cofolga2mo ' ;
   $rm = 'rm -rf';
   $PWD = 'pwd' ;
   $rnafold = 'RNAfold -p' ;
   $SLASH = '\/' ;
} else { # win
   # windows ms-dos
   $exe = "cofolga2mo.exe  " ;
   $rm = "del /Q /F";
   $PWD = "cd" ;
   $rnafold = "RNAfold.exe -p";
   $SLASH = "\\" ;
}

# temporary directory/file names
$tempdir1 = $ARGV[0] ;
$tempdir1 =~ s/^.*\/// ;
$tempdir1 = $tempdir1.".cofolga2mo.wrkdir.$$" ;
$tempfile1 = "cofolga.temp1" ; # fasta 
$tempfile2 = "cofolga.temp2" ; # prob. mat.

$root = `$PWD` ;
$root =~ s/\n$// ;

my @sq ;

open(IN,$ARGV[0]);

mkdir($tempdir1) ;
chdir($tempdir1) ;

# -f [fasta file]
# -pmf [BPP matrix file]
# -outdiv  -> output all individuals
# -pop [GA population size]
# -it [maximum GA iteration number]
# -outint [interavl for output, 0 does not output except for the final result]
# -unique  ->  not allow duplicated individuals in one GA population
# -r [initial random nubmer]
#
my $pop = 50 ;
$options = "-f $tempfile1 -pmf $tempfile2 -outdiv -pop $pop  -outint 0 -unique -r 12345 ".join(' ', @ARGV[1..$#ARGV]) ;

my $sort = 'I' ; ;
my $alnum = 10 ; # max. outputted alignments 
for (my $i = 0 ; $i < scalar(@ARGV) ; $i++) {
   if ($ARGV[$i] eq '--I') { $sort = 'I' ; }
   if ($ARGV[$i] eq '--P') { $sort = 'P' ; }
   if ($ARGV[$i] eq '--s') { $sort = 's' ; }
   if ($ARGV[$i] eq '--n') { 
      if ($ARGV[$i+1] == 0) { $alnum = $pop ; }
      else {
         if ($ARGV[$i+1] < 1 and $ARGV[$i+1] > $pop) {
            die ("ERROR: --n %s is too small or too large.\n", $ARGV[$i+1] );
         }
         $alnum = $ARGV[$i+1] ;
      }
   }
}

# delete pre-existing RNAfold output files
system("$rm \*_dp.ps");

# the follwings are necessary for the correct output order of sequences.
my $i = 1 ; my @id = (); my $inp = '' ;
while(<IN>) {
   chomp;
   if (/^>/) {
      if ($i != 1) {
         $inp .= "\n" ;
      }
      $id[$i] = $_ ;
      $id[$i] =~ s/^>\s*// ;
      $inp .= sprintf("> %03d\n",$i) ;
      $i++ ;

   } else {
      # convert a letter othr than ACGUT to N
      $_ =~ tr/a-z/A-Z/ ;
      $_ =~ s/[^ACGTU]/N/g ;
      $inp .= $_ ;
   }
}
$inp .= "\n" ;
my $nsq = $i - 1 ;

&run_RNAfold($inp) ;

# set input
open(OUT,">$tempfile2") ;
for ($i = 0 ; $i < $nsq ; $i++) { 
   &write_pmat($i) ;
}
close(OUT);

&run_exe ($options) ; 

&postprocess() ;


sub postprocess() {
   chdir($root) ;
   system("$rm $tempdir1");
}


sub run_RNAfold {
   my (@ps) ;
   open(OUT,">$tempfile1") ;
   print OUT $_[0];
   close(OUT);
   
   `$rnafold < $tempfile1` ;
   
   @ps = glob("*_dp.ps") ;
   &set_pmat_and_fa(@ps) ;
   system("$rm *_dp.ps *_ss.ps") ; #  do not delete $tempfile1
}


sub set_pmat_and_fa {
# derived from pmat.pl
# input: @_ (@files)
# output: (@fasta, @pmat, @fastaline, $ntot, @pmatline) (global)
   my ($i, $j, $f, @pm_a, @pm_b, @pm_c, $sw, $sq, @a, $num) ;
   my @files = @_ ;

   @fasta = () ; @pmat = () ;

   for($i = 0 ; $i <= $#files ; $i++) {
      $f = $files[$i] ;

      @pm_a = () ;@pm_b = () ;@pm_c = () ;
      $sw = 0 ;
      open(IN,"$f") ;
      while(<IN>) {
         chomp;
         if(/^\/sequence/) {
            $sq = "" ;
            while(<IN>) {
               if (/^\) \} def/) {last;}
               chomp ; 
               $_ =~ s/.$// ;  # delete \
               $sq .= $_ ;
            }

         } elsif (/^showpage/ and $sw == 1) {
            last ;
         } elsif (/^\%data starts here/) {
            $sw = 1 ;
         } elsif ($sw == 1 and /ubox$/) {
            $_ =~ s/ .box$// ;
            @a = () ; @a = split ;
            push(@pm_a, $a[0]) ;
            push(@pm_b, $a[1]) ;
            push(@pm_c, $a[2]*$a[2]) ; # because, $a[2] = sqrt(pij).
                                       # see _dp.ps file.
         }
      }
      close(IN) ;
      
      $id = $i+1 ;
      $fasta[$i][0] = sprintf("> %03d\n", $id) ;
      $fasta[$i][1] = "$sq\n" ;

      $num = $#pm_a + 1 ;
      $pmat[$i][0] = sprintf("> %d %03d\n", $num, $id) ;
      for ($j = 0 ; $j <= $#pm_a ; $j++) {
         $pmat[$i][$j+1] = "$pm_a[$j] $pm_b[$j] $pm_c[$j]\n" ;
      }
      $pmatline[$i] = $num + 1 ;

   }
}


sub run_exe {
# Input: $opt, $inp(local), $tempfile1, $pmat_name, $fname
# Output: @output (local), stdout
   my ($opt) = @_ ;

   @output = `$exe $opt` ;

   #print @output ;
   &linear_index() ;
}


sub write_pmat {
# Input: (@fasta, @pmat, @fastaline, @pmatline) (global)
#        $i, $fname, $pmat_name
# Output: $inp
   my ($i) = @_ ;

   for (my $k = 0 ; $k < $pmatline[$i] ; $k++) {
      print OUT  $pmat[$i][$k] ;
   }
}


sub linear_index {

chomp(@output) ;
my $ns = -1 ;
@sq = ();
$sq[0][0] = '' ;
my @a = () ;


# find alignments with the highest s
my $high_s = -999999999.0 ;
my $high_i ;
for( my $i = 0 ; $i <= $#output ; $i++) {
   $_ = $output[$i] ;
   if (/^\s*$/) { next ; }
   if (/LOOP/) { next ; }
   if (/  Individual/) {
      @a = split ;
      if ($a[3] == 1) { # rank 1
           $ns++ ;
	   if ($a[5] > $high_s) { $high_s = $a[5] ; $high_i = $ns ; }
      }
   } elsif (/> 1/) {
       if ($a[3] == 1) { # rank 1
	   $sq[0][$ns] = $output[$i+1] ;
       }
    } elsif (/> 2/) {
       if ($a[3] == 1) { # rank 1
	   $sq[1][$ns] = $output[$i+1] ;
	   #if ($ns == $alnum-1) { last ;} 
       }
   }
}

# seqid
my $match = 0 ;
my $tot = 0 ;

for (my $i = 0 ; $i < length($sq[0][$high_i]) ; $i++) {
   my $x = substr($sq[0][$high_i],$i,1) ;
   my $y = substr($sq[1][$high_i],$i,1) ;
   $x =~ tr/a-z/A-Z/ ;
   $y =~ tr/a-z/A-Z/ ;
  if($x eq $y and $x ne '-' ) { 
       $match++ ;
   } 
   if($x ne '-' and $y ne '-') { $tot++ ; } 
}
my $pid = 100.0*$match/$tot ; 

#print "#pid $pid\n" ;

my @dat = ();
$ns = -1 ; # reset 
@a = () ; #reset

if ($sort eq 's') { $sorti = 5 ; } 
elsif ($sort eq 'P') { $sorti = 6 ; }
elsif ($sort eq 'I') { $sorti = 7 ; }

my @sortdat = () ;
my @sortind = () ;

for( my $i = 0 ; $i <= $#output ; $i++) {
   $_ = $output[$i] ;
   if (/^\s*$/) { next ; }
   if (/LOOP/) { next ; }
   if (/  Individual/) {
      @a = split ;
      if ($a[3] == 1) { # rank 1
           $ns++ ;
	    #my $ms = $ns + 1 ;
	    # push(@a,-$ms);
	    # linear  $w = -$par1*$pid[$q] + $par2
	    my $w = -0.4*$pid + 40.0 ;
	    my $indx = $a[5] + $w*$a[6] ;   
            push(@a,$indx);
            push(@dat,  sprintf("%.2f %.2f %.2f",$indx,$a[5],$a[6])) ;

         # sort index
         my $j = $ns - 1 ;
         for ( ; $j >= 0  ; $j-- ) {
             if ($a[$sorti] > $sortdat[$j]) {
	        $sortdat[$j+1] = $sortdat[$j] ;
	        $sortind[$j+1] = $sortind[$j] ;
	     } else { last ; }
         }
         $sortdat[$j+1] = $a[$sorti] ;
         $sortind[$j+1] = $ns ;

	 #if ($ns == $alnum-1) { last ;} 
      }
   }
}

# $alnum is a max # output aligments
if ($alnum < $ns) { $ns = $alnum ; }
printf("# Alignments sorted by %s (only those with rank 1 are output)\n",$sort
);
printf("# I: linear weight index, P: structure score, s: sequence alignment score\n");
for (my $i = 0 ; $i < $ns ; $i++) {
   my @a = split(/\s+/,$dat[$sortind[$i]]) ;
   printf("  No. %d  I,P,s= %s %s %s\n\n",$i+1,$a[0],$a[2], $a[1]) ;
   printf("> $id[1]\n");
   printf("$sq[0][$sortind[$i]]\n");
   printf("> $id[2]\n");
   printf("$sq[1][$sortind[$i]]\n");
}

}

